@@ -10,9 +10,12 @@
 #include "elaborate.h"
 #include "parse.h"
 #include "tapsets.h"
+#include "session.h"
+#include "util.h"
 
 extern "C" {
 #include <sys/utsname.h>
+#include <fnmatch.h>
 }
 
 #include <algorithm>
@@ -28,27 +31,12 @@ extern "C" {
 using namespace std;
 
 
-template <typename OUT, typename IN> inline OUT
-lex_cast(IN const & in)
-{
-  stringstream ss;
-  OUT out;
-  if (!(ss << in && ss >> out))
-    throw runtime_error("bad lexical cast");
-  return out;
-}
-
-
 // ------------------------------------------------------------------------
 
 
-unsigned derived_probe::last_probeidx = 0;
-
 derived_probe::derived_probe (probe *p):
   base (p)
 {
-  this->name = string ("probe_") + lex_cast<string>(last_probeidx ++);
-
   if (p)
     {
       this->locations = p->locations;  
@@ -61,8 +49,6 @@ derived_probe::derived_probe (probe *p):
 derived_probe::derived_probe (probe *p, probe_point *l):
   base (p)
 {
-  this->name = string ("probe_") + lex_cast<string>(last_probeidx ++);
-
   if (l)
     this->locations.push_back (l);
 
@@ -74,6 +60,19 @@ derived_probe::derived_probe (probe *p, probe_point *l):
 }
 
 
+probe_point*
+derived_probe::sole_location ()
+{
+  if (locations.size() == 0)
+    throw semantic_error ("derived_probe with no locations", this->tok);
+  else if (locations.size() > 1)
+    throw semantic_error ("derived_probe with too many locations", this->tok);
+  else 
+    return locations[0];
+}
+
+
+
 // ------------------------------------------------------------------------
 // Members of derived_probe_builder
 
@@ -171,6 +170,23 @@ match_key::operator<(match_key const & other) const
 	      && parameter_type < other.parameter_type));
 }
 
+static bool
+isglob(string const & str)
+{
+  return(str.find('*') != str.npos);
+}
+
+bool
+match_key::globmatch(match_key const & other) const
+{
+  const char *other_str = other.name.c_str();
+  const char *name_str = name.c_str();
+
+  return ((fnmatch(name_str, other_str, FNM_NOESCAPE) == 0)
+	  && have_parameter == other.have_parameter 
+	  && parameter_type == other.parameter_type);
+}
+
 // ------------------------------------------------------------------------
 // Members of match_node
 // ------------------------------------------------------------------------
@@ -248,37 +264,45 @@ match_node::find_and_build (systemtap_session& s,
 
       b->build (s, p, loc, param_map, results);
     }
-  else if (loc->components[pos]->functor == "*") // wildcard?
+  else if (isglob(loc->components[pos]->functor)) // wildcard?
     {
-      // Recursively call derive_probes for all matches of the current
-      // key position.  To do this, we have to perform almost an
-      // alias-level duplication of the probe body and synthesis of
-      // new probe_points.
+      match_key match (* loc->components[pos]);
 
-      probe * n = new probe();
-      n->tok = p->tok;
-      n->body = deep_copy_visitor::deep_copy(p->body);
-
-      // Construct N probe_point instances: one per wildcard match
+      // Call find_and_build for each possible match.  Ignore errors -
+      // unless we don't find any match.
+      unsigned int num_results = results.size();
       for (sub_map_iterator_t i = sub.begin(); i != sub.end(); i++)
         {
-          const match_key& match = i->first;
-          probe_point *loc2 = new probe_point;
-          loc2->tok = loc->tok;
+	  const match_key& subkey = i->first;
+	  match_node* subnode = i->second;
 
-          // deep-copy probe point components, except for this wildcard position
-          for (unsigned k=0; k<loc->components.size(); k++)
-            if (pos != k)
-              loc2->components.push_back(new probe_point::component(loc->components[k]->functor,
-                                                                    loc->components[k]->arg));
-            else
-              // NB: we retain wildcard arg - foo.*(5).bar => foo.a(5).bar, foo.b(5).bar
-              loc2->components.push_back(new probe_point::component(match.name,
-                                                                    loc->components[k]->arg));
-          n->locations.push_back (loc2);
-        }
-
-      derive_probes (s, n, results, false);
+	  if (match.globmatch(subkey))
+	    {
+	      // recurse
+	      try
+	        {
+		  subnode->find_and_build (s, p, loc, pos+1, results);
+	        }
+	      catch (const semantic_error& e)
+	        {
+		  // Ignore semantic_errors while expanding wildcards.
+		  // If we get done and nothing was expanded, the code
+		  // following the loop will complain.
+		}
+	    }
+	}
+      if (! loc->optional && num_results == results.size())
+        {
+	  // We didn't find any wildcard matches (since the size of
+	  // the result vector didn't change).  Throw an error.
+          string alternatives;
+          for (sub_map_iterator_t i = sub.begin(); i != sub.end(); i++)
+            alternatives += string(" ") + i->first.str();
+          
+	  throw semantic_error(string("probe point mismatch at position ") +
+			       lex_cast<string> (pos) +
+			       " (alternatives:" + alternatives + ")");
+	}
     }
   else 
     {
@@ -306,6 +330,20 @@ match_node::find_and_build (systemtap_session& s,
 // Alias probes
 // ------------------------------------------------------------------------
 
+struct alias_derived_probe: public derived_probe
+{
+  alias_derived_probe (probe* base): derived_probe (base) {}
+
+  void upchuck () { throw semantic_error ("inappropriate", this->tok); }
+
+  // Alias probes are immediately expanded to other derived_probe
+  // types, and are not themselves emitted or listed in
+  // systemtap_session.probes
+
+  void join_group (systemtap_session& s) { upchuck (); }
+};
+
+
 struct
 alias_expansion_builder 
   : public derived_probe_builder
@@ -326,7 +364,7 @@ alias_expansion_builder
     // alias_expansion_probe so that the expansion loop recognizes it as
     // such and re-expands its expansion.
     
-    probe * n = new probe();
+    probe * n = new alias_derived_probe (use);
     n->body = new block();
 
     // The new probe gets the location list of the alias,
@@ -344,19 +382,28 @@ alias_expansion_builder
     // there's concatenated code here and we only want one vardecl per
     // resulting variable.
 
-    for (unsigned i = 0; i < alias->body->statements.size(); ++i)
+    if (alias->epilogue_style)
       {
-	statement *s = deep_copy_visitor::deep_copy(alias->body->statements[i]);
-	n->body->statements.push_back(s);
-      }
+        for (unsigned i = 0; i < use->body->statements.size(); ++i)
+          n->body->statements.push_back
+            (deep_copy_visitor::deep_copy(use->body->statements[i]));
 
-    for (unsigned i = 0; i < use->body->statements.size(); ++i)
+        for (unsigned i = 0; i < alias->body->statements.size(); ++i)
+          n->body->statements.push_back
+            (deep_copy_visitor::deep_copy(alias->body->statements[i]));
+      }
+    else
       {
-	statement *s = deep_copy_visitor::deep_copy(use->body->statements[i]);
-	n->body->statements.push_back(s);
+        for (unsigned i = 0; i < alias->body->statements.size(); ++i)
+          n->body->statements.push_back
+            (deep_copy_visitor::deep_copy(alias->body->statements[i]));
+        
+        for (unsigned i = 0; i < use->body->statements.size(); ++i)
+          n->body->statements.push_back
+            (deep_copy_visitor::deep_copy(use->body->statements[i]));
       }
-
-    derive_probes (sess, n, finished_results, false);
+    
+    derive_probes (sess, n, finished_results, location->optional);
   }
 };
 
@@ -402,10 +449,14 @@ systemtap_session::register_library_aliases()
             }
           catch (const semantic_error& e)
             {
-              cerr << "while: registering probe alias ";
-              alias->printsig(cerr);
-              cerr << endl;
-              print_error (e);
+              semantic_error* er = new semantic_error (e); // copy it
+              stringstream msg;
+              msg << e.msg2;
+              msg << " while registering probe alias ";
+              alias->printsig(msg);
+              er->msg2 = msg.str();
+              print_error (* er);
+              delete er;
             }
 	}
     }
@@ -434,29 +485,44 @@ recursion_guard
 void
 derive_probes (systemtap_session& s,
                probe *p, vector<derived_probe*>& dps,
-               bool exc_outermost)
+               bool optional)
 {
   for (unsigned i = 0; i < p->locations.size(); ++i)
     {
       probe_point *loc = p->locations[i];
-      
+
+      // Pass down optional flag from e.g. alias reference to each
+      // probe_point instance.  We do this by temporarily overriding
+      // the probe_point optional flag.  We could instead deep-copy
+      // and set a flag on the copy permanently.
+
+      bool old_loc_opt = loc->optional;
+      loc->optional = loc->optional || optional;
+
       try
         {
           unsigned num_atbegin = dps.size();
           s.pattern_root->find_and_build (s, p, loc, 0, dps);
           unsigned num_atend = dps.size();
           
-          if (num_atbegin == num_atend) // nothing new derived!
+          if (! loc->optional && // something required, but
+              num_atbegin == num_atend) // nothing new derived!
             throw semantic_error ("no match for probe point");
         }
       catch (const semantic_error& e)
         {
           // XXX: prefer not to print_error at every nest/unroll level
-          s.print_error (e);
-          cerr << "while: resolving probe point " << *loc << endl;
-          //          if (! exc_outermost)
-          //            throw;
+
+          semantic_error* er = new semantic_error (e); // copy it
+          stringstream msg;
+          msg << e.msg2;
+          msg << " while resolving probe point " << *loc;
+          er->msg2 = msg.str();
+          s.print_error (* er);
+          delete er;
         }
+
+      loc->optional = old_loc_opt;
     }
 }
 
@@ -653,7 +719,7 @@ semantic_pass_vars (systemtap_session & sess)
 	sess.probes[i]->body->visit (&chk);
     }  
 
-  return sess.num_errors;
+  return sess.num_errors();
 }
 
 // ------------------------------------------------------------------------
@@ -755,14 +821,15 @@ semantic_pass_stats (systemtap_session & sess)
 	}
     }
   
-  return sess.num_errors;
+  return sess.num_errors();
 }
 
 // ------------------------------------------------------------------------
 
 
 static int semantic_pass_symbols (systemtap_session&);
-static int semantic_pass_optimize (systemtap_session&);
+static int semantic_pass_optimize1 (systemtap_session&);
+static int semantic_pass_optimize2 (systemtap_session&);
 static int semantic_pass_types (systemtap_session&);
 static int semantic_pass_vars (systemtap_session&);
 static int semantic_pass_stats (systemtap_session&);
@@ -824,14 +891,15 @@ semantic_pass_symbols (systemtap_session& s)
           probe* p = dome->probes [i];
           vector<derived_probe*> dps;
 
-          // much magic happens here: probe alias expansion,
-          // provider identification
+          // much magic happens here: probe alias expansion, wildcard
+          // matching, low-level derived_probe construction.
           derive_probes (s, p, dps);
 
           for (unsigned j=0; j<dps.size(); j++)
             {
               derived_probe* dp = dps[j];
-	      s.probes.push_back (dp);
+              s.probes.push_back (dp);
+              dp->join_group (s);
 
               try 
                 {
@@ -847,7 +915,7 @@ semantic_pass_symbols (systemtap_session& s)
         }
     }
   
-  return s.num_errors; // all those print_error calls
+  return s.num_errors(); // all those print_error calls
 }
 
 
@@ -863,8 +931,9 @@ semantic_pass (systemtap_session& s)
       register_standard_tapsets(s);
       
       rc = semantic_pass_symbols (s);
-      if (rc == 0 && ! s.unoptimized) rc = semantic_pass_optimize (s);
+      if (rc == 0 && ! s.unoptimized) rc = semantic_pass_optimize1 (s);
       if (rc == 0) rc = semantic_pass_types (s);
+      if (rc == 0 && ! s.unoptimized) rc = semantic_pass_optimize2 (s);
       if (rc == 0) rc = semantic_pass_vars (s);
       if (rc == 0) rc = semantic_pass_stats (s);
     }
@@ -881,8 +950,19 @@ semantic_pass (systemtap_session& s)
 
 
 systemtap_session::systemtap_session ():
+  // NB: pointer members must be manually initialized!
   pattern_root(new match_node),
-  user_file (0), op (0), up (0), num_errors (0)
+  user_file (0),
+  be_derived_probes(0), 
+  dwarf_derived_probes(0), 
+  timer_derived_probes(0), 
+  profile_derived_probes(0), 
+  mark_derived_probes(0), 
+  hrtimer_derived_probes(0), 
+  perfmon_derived_probes(0), 
+  op (0), up (0),
+  kprobes_text_initialized (false),
+  kprobes_text_start (0), kprobes_text_end (0)
 {
 }
 
@@ -890,14 +970,27 @@ systemtap_session::systemtap_session ():
 void
 systemtap_session::print_error (const semantic_error& e)
 {
-  cerr << "semantic error: " << e.what ();
+  string message_str;
+  stringstream message;
+
+  message << "semantic error: " << e.what ();
   if (e.tok1 || e.tok2)
-    cerr << ": ";
-  if (e.tok1) cerr << *e.tok1;
-  cerr << e.msg2;
-  if (e.tok2) cerr << *e.tok2;
-  cerr << endl;
-  num_errors ++;
+    message << ": ";
+  if (e.tok1) message << *e.tok1;
+  message << e.msg2;
+  if (e.tok2) message << *e.tok2;
+  message << endl;
+  message_str = message.str();
+
+  // Duplicate elimination
+  if (seen_errors.find (message_str) == seen_errors.end())
+    {
+      seen_errors.insert (message_str);
+      cerr << message_str;
+    }
+
+  if (e.chain)
+    print_error (* e.chain);
 }
 
 
@@ -947,9 +1040,10 @@ symresolution_info::visit_foreach_loop (foreach_loop* e)
 	    array->referent = d;
 	  else
             {
-              cerr << "while searching for arity " << e->indexes.size()
-                   << " array:\n";
-              throw semantic_error ("unresolved global array " + array->name, e->tok);
+              stringstream msg;
+              msg << "unresolved arity-" << e->indexes.size()
+                  << " global array " << array->name;
+              throw semantic_error (msg.str(), e->tok);
             }
 	}
     }
@@ -959,6 +1053,9 @@ symresolution_info::visit_foreach_loop (foreach_loop* e)
       hist->visit (this);
     }
 
+  if (e->limit)
+    e->limit->visit (this);
+
   e->block->visit (this);
 }
 
@@ -1087,8 +1184,10 @@ symresolution_info::visit_functioncall (functioncall* e)
     e->referent = d;
   else
     {
-      cerr << "while searching for arity " << e->args.size() << " function:\n";
-      throw semantic_error ("unresolved function call", e->tok);
+      stringstream msg;
+      msg << "unresolved arity-" << e->args.size()
+          << " function";
+      throw semantic_error (msg.str(), e->tok);
     }
 }
 
@@ -1500,9 +1599,104 @@ void semantic_pass_opt4 (systemtap_session& s, bool& relaxed_p)
     }
 }
 
+struct duplicate_function_remover: public functioncall_traversing_visitor
+{
+  systemtap_session& s;
+  map<functiondecl*, functiondecl*>& duplicate_function_map;
+
+  duplicate_function_remover(systemtap_session& sess,
+			     map<functiondecl*, functiondecl*>&dfm):
+    s(sess), duplicate_function_map(dfm) {};
+
+  void visit_functioncall (functioncall* e);
+};
+
+void
+duplicate_function_remover::visit_functioncall (functioncall *e)
+{
+  functioncall_traversing_visitor::visit_functioncall (e);
+
+  // If the current function call reference points to a function that
+  // is a duplicate, replace it.
+  if (duplicate_function_map.count(e->referent) != 0)
+    {
+      if (s.verbose>2)
+	  clog << "Changing " << e->referent->name
+	       << " reference to "
+	       << duplicate_function_map[e->referent]->name
+	       << " reference\n";
+      e->tok = duplicate_function_map[e->referent]->tok;
+      e->function = duplicate_function_map[e->referent]->name;
+      e->referent = duplicate_function_map[e->referent];
+    }
+}
+
+static string
+get_functionsig (functiondecl* f)
+{
+  ostringstream s;
+
+  // Get the "name:args body" of the function in s.  We have to
+  // include the args since the function 'x1(a, b)' is different than
+  // the function 'x2(b, a)' even if the bodies of the two functions
+  // are exactly the same.
+  f->printsig(s);
+  f->body->print(s);
+
+  // printsig puts f->name + ':' on the front.  Remove this
+  // (otherwise, functions would never compare equal).
+  string str = s.str().erase(0, f->name.size() + 1);
+
+  // Return the function signature.
+  return str;
+}
+
+void semantic_pass_opt5 (systemtap_session& s, bool& relaxed_p)
+{
+  // Walk through all the functions, looking for duplicates.
+  map<string, functiondecl*> functionsig_map;
+  map<functiondecl*, functiondecl*> duplicate_function_map;
+  for (unsigned i=0; i < s.functions.size(); /* see below */)
+    {
+      string functionsig = get_functionsig(s.functions[i]);
+
+      if (functionsig_map.count(functionsig) == 0)
+	{
+	  // This function is unique.  Remember it.
+	  functionsig_map[functionsig] = s.functions[i];
+	  i++;
+	}
+      else
+        {
+	  // This function is a duplicate.
+	  duplicate_function_map[s.functions[i]]
+	    = functionsig_map[functionsig];
+
+	  // Remove the duplicate function (since we don't need it
+	  // anymore).
+	  s.functions.erase (s.functions.begin() + i);
+
+	  relaxed_p = false;
+          // NB: don't increment i
+	}
+    }
+
+  // If we have duplicate functions, traverse down the tree, replacing
+  // the appropriate function calls.
+  // duplicate_function_remover::visit_functioncall() handles the
+  // details of replacing the function calls.
+  if (duplicate_function_map.size() != 0)
+    {
+      duplicate_function_remover dfr (s, duplicate_function_map);
+
+      for (unsigned i=0; i < s.probes.size(); i++)
+	s.probes[i]->body->visit(&dfr);
+    }
+}
+
 
 static int
-semantic_pass_optimize (systemtap_session& s)
+semantic_pass_optimize1 (systemtap_session& s)
 {
   // In this pass, we attempt to rewrite probe/function bodies to
   // eliminate some blatantly unnecessary code.  This is run before
@@ -1533,6 +1727,33 @@ semantic_pass_optimize (systemtap_session& s)
 }
 
 
+static int
+semantic_pass_optimize2 (systemtap_session& s)
+{
+  // This is run after type inference.  We run an outer "relaxation"
+  // loop that repeats the optimizations until none of them find
+  // anything to remove.
+
+  int rc = 0;
+
+  bool relaxed_p = false;
+  while (! relaxed_p)
+    {
+      relaxed_p = true; // until proven otherwise
+
+      semantic_pass_opt5 (s, relaxed_p);
+    }
+
+  if (s.probes.size() == 0)
+    {
+      cerr << "semantic error: no probes found." << endl;
+      rc = 1;
+    }
+
+  return rc;
+}
+
+
 
 // ------------------------------------------------------------------------
 // type resolution
@@ -1599,7 +1820,7 @@ semantic_pass_types (systemtap_session& s)
           }
     }
   
-  return rc + s.num_errors;
+  return rc + s.num_errors();
 }
 
 
@@ -1944,7 +2165,10 @@ typeresolution_info::visit_target_symbol (target_symbol* e)
   // later unused-expression-elimination pass didn't get rid of it
   // either.  So we have a target symbol that is believed to be of
   // genuine use, yet unresolved by the provider.
-  throw semantic_error("unresolved target-symbol expression", e->tok);
+  if (e->saved_conversion_error)
+    throw (* (e->saved_conversion_error));
+  else
+    throw semantic_error("unresolved target-symbol expression", e->tok);
 }
 
 
@@ -2169,6 +2393,12 @@ typeresolution_info::visit_foreach_loop (foreach_loop* e)
 	}
     }
 
+  if (e->limit)
+    {
+      t = pe_long;
+      e->limit->visit (this);
+    }
+
   t = pe_unknown;
   e->block->visit (this);  
 }
