@@ -15,6 +15,9 @@
 #include "translate.h"
 #include "buildrun.h"
 #include "session.h"
+#include "hash.h"
+#include "cache.h"
+#include "util.h"
 
 #include <iostream>
 #include <fstream>
@@ -28,6 +31,7 @@ extern "C" {
 #include <sys/utsname.h>
 #include <sys/times.h>
 #include <sys/time.h>
+#include <sys/stat.h>
 #include <time.h>
 #include <elfutils/libdwfl.h>
 }
@@ -101,17 +105,78 @@ usage (systemtap_session& s, int exitcode)
 }
 
 
-// little utility function
-
-template <typename T>
-static string
-stringify(T t)
+static void
+printscript(systemtap_session& s, ostream& o)
 {
-  ostringstream s;
-  s << t;
-  return s.str ();
-}
+  if (s.embeds.size() > 0)
+    o << "# global embedded code" << endl;
+  for (unsigned i=0; i<s.embeds.size(); i++)
+    {
+      embeddedcode* ec = s.embeds[i];
+      ec->print (o);
+      o << endl;
+    }
 
+  if (s.globals.size() > 0)
+    o << "# globals" << endl;
+  for (unsigned i=0; i<s.globals.size(); i++)
+    {
+      vardecl* v = s.globals[i];
+      v->printsig (o);
+      if (s.verbose && v->init)
+        {
+          o << " = ";
+          v->init->print(o);
+        }
+      o << endl;
+    }
+
+  if (s.functions.size() > 0)
+    o << "# functions" << endl;
+  for (unsigned i=0; i<s.functions.size(); i++)
+    {
+      functiondecl* f = s.functions[i];
+      f->printsig (o);
+      o << endl;
+      if (f->locals.size() > 0)
+	o << "  # locals" << endl;
+      for (unsigned j=0; j<f->locals.size(); j++)
+        {
+	  vardecl* v = f->locals[j];
+	  o << "  ";
+	  v->printsig (o);
+	  o << endl;
+	}
+      if (s.verbose)
+        {
+	  f->body->print (o);
+	  o << endl;
+	}
+    }
+
+  if (s.probes.size() > 0)
+    o << "# probes" << endl;
+  for (unsigned i=0; i<s.probes.size(); i++)
+    {
+      derived_probe* p = s.probes[i];
+      p->printsig (o);
+      o << endl;
+      if (p->locals.size() > 0)
+        o << "  # locals" << endl;
+      for (unsigned j=0; j<p->locals.size(); j++)
+        {
+	  vardecl* v = p->locals[j];
+	  o << "  ";
+	  v->printsig (o);
+	  o << endl;
+	}
+      if (s.verbose)
+        {
+	  p->body->print (o);
+	  o << endl;
+	}
+    }
+}
 
 int
 main (int argc, char * const argv [])
@@ -119,6 +184,7 @@ main (int argc, char * const argv [])
   string cmdline_script; // -e PROGRAM
   string script_file; // FILE
   bool have_script = false;
+  bool release_changed = false;
 
   // Initialize defaults
   systemtap_session s;
@@ -140,6 +206,8 @@ main (int argc, char * const argv [])
   s.target_pid = 0;
   s.merge=true;
   s.perfmon=0;
+  s.symtab = false;
+  s.use_cache = true;
 
   const char* s_p = getenv ("SYSTEMTAP_TAPSET");
   if (s_p != NULL)  
@@ -159,6 +227,33 @@ main (int argc, char * const argv [])
   else
     s.runtime_path = string(PKGDATADIR) + "/runtime";
 
+  const char* s_d = getenv ("SYSTEMTAP_DIR");
+  if (s_d != NULL)
+    s.data_path = s_d;
+  else
+    s.data_path = get_home_directory() + string("/.systemtap");
+  if (create_dir(s.data_path.c_str()) == 1)
+    {
+      const char* e = strerror (errno);
+      cerr << "Warning: failed to create systemtap data directory (\""
+	   << s.data_path << "\"): " << e << endl;
+      cerr << "Disabling cache support." << endl;
+      s.use_cache = false;
+    }
+
+  if (s.use_cache)
+    {
+      s.cache_path = s.data_path + "/cache";
+      if (create_dir(s.cache_path.c_str()) == 1)
+        {
+	  const char* e = strerror (errno);
+	  cerr << "Warning: failed to create cache directory (\""
+	       << s.cache_path << "\"): " << e << endl;
+	  cerr << "Disabling cache support." << endl;
+	  s.use_cache = false;
+	}
+    }
+
   while (true)
     {
       int grc = getopt (argc, argv, "hVMvtp:I:e:o:R:r:m:kgc:x:D:bs:u");
@@ -216,10 +311,13 @@ main (int argc, char * const argv [])
 
         case 'm':
           s.module_name = string (optarg);
+	  cerr << "Warning: using '-m' disables cache support." << endl;
+	  s.use_cache = false;
           break;
 
         case 'r':
           s.kernel_release = string (optarg);
+	  release_changed = true;
           break;
 
         case 'k':
@@ -281,6 +379,13 @@ main (int argc, char * const argv [])
       usage (s, 1);
     }
 
+  if (s.last_pass > 4 && release_changed)
+    {
+      cerr << ("Warning: changing last pass to 4 since the kernel release"
+	       " has changed.") << endl;
+      s.last_pass = 4;
+    }
+
   for (int i = optind; i < argc; i++)
     {
       if (! have_script)
@@ -313,8 +418,9 @@ main (int argc, char * const argv [])
 
   s.kernel_base_release.assign(s.kernel_release, 0, s.kernel_release.find('-'));
 
-
   // arguments parsed; get down to business
+  if (s.verbose > 1)
+    version ();
 
   // Create a temporary directory to build within.
   // Be careful with this, as "s.tmpdir" is "rm -rf"'d at the end.
@@ -339,18 +445,30 @@ main (int argc, char * const argv [])
       clog << "Created temporary directory \"" << s.tmpdir << "\"" << endl;
   }
 
+  // Create the name of the C source file within the temporary
+  // directory.
+  s.translated_source = string(s.tmpdir) + "/" + s.module_name + ".c";
+
   struct tms tms_before;
   times (& tms_before);
   struct timeval tv_before;
   gettimeofday (&tv_before, NULL);
 
   // PASS 1a: PARSING USER SCRIPT
-  // XXX: pass args vector, so parser (or lexer?) can substitute
-  // $1..$NN with actual arguments
+
+  struct stat user_file_stat;
+  int user_file_stat_rc = -1;
+
   if (script_file == "-")
-    s.user_file = parser::parse (s, cin, s.guru_mode);
+    {
+      s.user_file = parser::parse (s, cin, s.guru_mode);
+      user_file_stat_rc = fstat (STDIN_FILENO, & user_file_stat);
+    }
   else if (script_file != "")
-    s.user_file = parser::parse (s, script_file, s.guru_mode);
+    {
+      s.user_file = parser::parse (s, script_file, s.guru_mode);
+      user_file_stat_rc = stat (script_file.c_str(), & user_file_stat);
+    }
   else
     {
       istringstream ii (cmdline_script);
@@ -407,11 +525,24 @@ main (int argc, char * const argv [])
           for (unsigned j=0; j<globbuf.gl_pathc; j++)
             {
               // privilege only for /usr/share/systemtap?
+              
               stapfile* f = parser::parse (s, globbuf.gl_pathv[j], true);
               if (f == 0)
                 rc ++;
               else
                 s.library_files.push_back (f);
+
+              struct stat tapset_file_stat;
+              int stat_rc = stat (globbuf.gl_pathv[j], & tapset_file_stat);
+              if (stat_rc == 0 && user_file_stat_rc == 0 &&
+                  user_file_stat.st_dev == tapset_file_stat.st_dev &&
+                  user_file_stat.st_ino == tapset_file_stat.st_ino)
+                {
+                  clog << "usage error: tapset file '" << globbuf.gl_pathv[j]
+                       << "' cannot be run directly as a session script." << endl;
+                  rc ++;
+                }
+
             }
 
           globfree (& globbuf);
@@ -469,62 +600,7 @@ main (int argc, char * const argv [])
   rc = semantic_pass (s);
 
   if (rc == 0 && s.last_pass == 2)
-    {
-      if (s.globals.size() > 0)
-        cout << "# globals" << endl;
-      for (unsigned i=0; i<s.globals.size(); i++)
-	{
-	  vardecl* v = s.globals[i];
-	  v->printsig (cout);
-          cout << endl;
-	}
-
-      if (s.functions.size() > 0)
-        cout << "# functions" << endl;
-      for (unsigned i=0; i<s.functions.size(); i++)
-	{
-	  functiondecl* f = s.functions[i];
-	  f->printsig (cout);
-          cout << endl;
-          if (f->locals.size() > 0)
-            cout << "  # locals" << endl;
-          for (unsigned j=0; j<f->locals.size(); j++)
-            {
-              vardecl* v = f->locals[j];
-              cout << "  ";
-              v->printsig (cout);
-              cout << endl;
-            }
-          if (s.verbose)
-            {
-              f->body->print (cout);
-              cout << endl;
-            }
-	}
-
-      if (s.probes.size() > 0)
-        cout << "# probes" << endl;
-      for (unsigned i=0; i<s.probes.size(); i++)
-	{
-	  derived_probe* p = s.probes[i];
-	  p->printsig (cout);
-          cout << endl;
-          if (p->locals.size() > 0)
-            cout << "  # locals" << endl;
-          for (unsigned j=0; j<p->locals.size(); j++)
-            {
-              vardecl* v = p->locals[j];
-              cout << "  ";
-              v->printsig (cout);
-              cout << endl;
-            }
-          if (s.verbose)
-            {
-              p->body->print (cout);
-              cout << endl;
-            }
-	}
-    }
+    printscript(s, cout);
 
   times (& tms_after);
   gettimeofday (&tv_after, NULL);
@@ -532,6 +608,7 @@ main (int argc, char * const argv [])
   if (s.verbose) clog << "Pass 2: analyzed script: "
                       << s.probes.size() << " probe(s), "
                       << s.functions.size() << " function(s), "
+                      << s.embeds.size() << " embed(s), "
                       << s.globals.size() << " global(s) in "
                       << TIMESPRINT
                       << endl;
@@ -540,6 +617,38 @@ main (int argc, char * const argv [])
     cerr << "Pass 2: analysis failed.  "
          << "Try again with more '-v' (verbose) options."
          << endl;
+  // Generate hash.  There isn't any point in generating the hash
+  // if last_pass is 2, since we'll quit before using it.
+  else if (s.last_pass != 2 && s.use_cache)
+    {
+      ostringstream o;
+      unsigned saved_verbose;
+	  
+      // Make sure we're in verbose mode, so that printscript()
+      // will output function/probe bodies.
+      saved_verbose = s.verbose;
+      s.verbose = 3;
+
+      // Print script to 'o'
+      printscript(s, o);
+
+      // Restore original verbose mode setting.
+      s.verbose = saved_verbose;
+
+      // Generate hash
+      find_hash (s, o.str());
+
+      // See if we can use cached source/module.
+      if (get_from_cache(s))
+        {
+	  // If our last pass isn't 5, we're done (since passes 3 and
+	  // 4 just generate what we just pulled out of the cache).
+	  if (s.last_pass < 5) goto cleanup;
+
+	  // Short-circuit to pass 5.
+	  goto pass_5;
+	}
+    }
 
   if (rc || s.last_pass == 2) goto cleanup;
 
@@ -548,7 +657,6 @@ main (int argc, char * const argv [])
   times (& tms_before);
   gettimeofday (&tv_before, NULL);
 
-  s.translated_source = string(s.tmpdir) + "/" + s.module_name + ".c";
   rc = translate_pass (s);
 
   if (rc == 0 && s.last_pass == 3)
@@ -590,11 +698,17 @@ main (int argc, char * const argv [])
     cerr << "Pass 4: compilation failed.  "
          << "Try again with more '-v' (verbose) options."
          << endl;
+  else if (s.use_cache)
+    {
+      // Update cache.
+      add_to_cache(s);
+    }
 
   // XXX: what to do if rc==0 && last_pass == 4?  dump .ko file to stdout?
   if (rc || s.last_pass == 4) goto cleanup;
 
   // PASS 5: RUN
+pass_5:
   times (& tms_before);
   gettimeofday (&tv_before, NULL);
   // NB: this message is a judgement call.  The other passes don't emit
