@@ -162,11 +162,7 @@ bool eval_pp_conditional (systemtap_session& s,
                                     l->content == "kernel_vr"))
     {
       string target_kernel_vr = s.kernel_release;
-      string target_kernel_v = target_kernel_vr;
-      // cut off any release code suffix
-      string::size_type dr = target_kernel_vr.find ('-');
-      if (dr > 0 && dr != string::npos)
-        target_kernel_v = target_kernel_vr.substr (0, dr);
+      string target_kernel_v = s.kernel_base_release;
       
       if (! (r->type == tok_string))
         throw parse_error ("expected string literal", r);
@@ -530,7 +526,7 @@ lexer::scan ()
               // Use @1/$1 as the base, not @0/$0.  Thus the idx-1.
               if (errno == ERANGE || errno == EINVAL || *endp != '\0' ||
                   idx == 0 || idx-1 >= session.args.size ())
-                throw parse_error ("command line argument index invalid or out of range");
+                throw parse_error ("command line argument index invalid or out of range", n);
               
               string arg = session.args[idx-1];
               n->type = (n->content[0] == '@') ? tok_string : tok_number;
@@ -547,6 +543,7 @@ lexer::scan ()
 	      || n->content == "for"
 	      || n->content == "foreach"
 	      || n->content == "in"
+	      || n->content == "limit"
 	      || n->content == "return"
 	      || n->content == "delete"
 	      || n->content == "while"
@@ -785,7 +782,7 @@ parser::parse ()
 	  else if (t->type == tok_keyword && t->content == "global")
 	    {
 	      context = con_global;
-	      parse_global (f->globals);
+	      parse_global (f->globals, f->probes);
 	    }
 	  else if (t->type == tok_keyword && t->content == "function")
 	    {
@@ -1024,7 +1021,7 @@ parser::parse_statement ()
 
 
 void
-parser::parse_global (vector <vardecl*>& globals)
+parser::parse_global (vector <vardecl*>& globals, vector<probe*>& probes)
 {
   const token* t0 = next ();
   if (! (t0->type == tok_keyword && t0->content == "global"))
@@ -1046,7 +1043,31 @@ parser::parse_global (vector <vardecl*>& globals)
       globals.push_back (d);
 
       t = peek ();
-      if (t && t->type == tok_operator && t->content == ",")
+
+      if (t && t->type == tok_operator && t->content == "[") // array size
+	{
+	  int64_t size;
+	  next ();
+	  expect_number(size);
+	  if (size <= 0 || size > 1000000) // arbitrary max
+	    throw parse_error("array size out of range");
+	  d->maxsize = (int)size;
+	  expect_known(tok_operator, "]");
+	  t = peek ();
+	}
+
+      if (t && t->type == tok_operator && t->content == "=") // initialization
+	{
+	  if (!d->compatible_arity(0))
+	    throw parse_error("only scalar globals can be initialized");
+	  d->set_arity(0);
+	  next ();
+	  d->init = parse_literal ();
+	  d->type = d->init->type;
+	  t = peek ();
+	}
+
+      if (t && t->type == tok_operator && t->content == ",") // next global
 	{
 	  next ();
 	  continue;
@@ -1231,25 +1252,40 @@ parser::parse_literal ()
   literal* l;
   if (t->type == tok_string)
     l = new literal_string (t->content);
-  else if (t->type == tok_number)
-    {
-      const char* startp = t->content.c_str ();
-      char* endp = (char*) startp;
-
-      // NB: we allow controlled overflow from LLONG_MIN .. ULLONG_MAX
-      // Actually, this allows all the way from -ULLONG_MAX to ULLONG_MAX,
-      // since the lexer only gives us positive digit strings.
-      errno = 0;
-      long long value = (long long) strtoull (startp, & endp, 0);
-      if (errno == ERANGE || errno == EINVAL || *endp != '\0'
-          || (unsigned long long) value > 18446744073709551615ULL
-          || value < -9223372036854775807LL-1)
-        throw parse_error ("number invalid or out of range"); 
-
-      l = new literal_number (value);
-    }
   else
-    throw parse_error ("expected literal string or number");
+    {
+      bool neg = false;
+      if (t->type == tok_operator && t->content == "-")
+	{
+	  neg = true;
+	  t = next ();
+	}
+
+      if (t->type == tok_number)
+	{
+	  const char* startp = t->content.c_str ();
+	  char* endp = (char*) startp;
+
+	  // NB: we allow controlled overflow from LLONG_MIN .. ULLONG_MAX
+	  // Actually, this allows all the way from -ULLONG_MAX to ULLONG_MAX,
+	  // since the lexer only gives us positive digit strings, but we'll
+	  // limit it to LLONG_MIN when a '-' operator is fed into the literal.
+	  errno = 0;
+	  long long value = (long long) strtoull (startp, & endp, 0);
+	  if (errno == ERANGE || errno == EINVAL || *endp != '\0'
+	      || (neg && (unsigned long long) value > 9223372036854775808ULL)
+	      || (unsigned long long) value > 18446744073709551615ULL
+	      || value < -9223372036854775807LL-1)
+	    throw parse_error ("number invalid or out of range"); 
+
+	  if (neg)
+	    value = -value;
+
+	  l = new literal_number (value);
+	}
+      else
+	throw parse_error ("expected literal string or number");
+    }
 
   l->tok = t;
   return l;
@@ -1473,6 +1509,7 @@ parser::parse_foreach_loop ()
   foreach_loop* s = new foreach_loop;
   s->tok = t;
   s->sort_direction = 0;
+  s->limit = NULL;
 
   t = next ();
   if (! (t->type == tok_operator && t->content == "("))
@@ -1546,6 +1583,13 @@ parser::parse_foreach_loop ()
       next();
     }
 
+  t = peek ();
+  if (tok_is(t, tok_keyword, "limit"))
+    {
+      next ();				// get past the "limit"
+      s->limit = parse_expression ();
+    }
+
   t = next ();
   if (! (t->type == tok_operator && t->content == ")"))
     throw parse_error ("expected ')'");
