@@ -19,17 +19,9 @@
  * Copyright (C) Red Hat Inc, 2005, 2006
  *
  */
-
-#include <ctype.h>
-#include <stdio.h>
-#include <stdlib.h>
-#include <unistd.h>
-#include <string.h>
+#include "librelay.h"
 #include <signal.h>
-#include <sys/types.h>
-#include <sys/stat.h>
 #include <sys/ioctl.h>
-#include <fcntl.h>
 #include <errno.h>
 #include <linux/fd.h>
 #include <sys/mman.h>
@@ -40,8 +32,7 @@
 #include <linux/limits.h>
 #include <sys/wait.h>
 #include <sys/statfs.h>
-#include <stdint.h>
-#include "librelay.h"
+
 
 /* stp_check script */
 #ifdef PKGLIBDIR
@@ -87,7 +78,7 @@ static char *relay_buffer[NR_CPUS];
 static pthread_t reader[NR_CPUS];
 
 /* control channel */
-static int control_channel;
+int control_channel;
 
 /* flags */
 extern int print_only, quiet, verbose;
@@ -106,10 +97,12 @@ extern gid_t cmd_gid;
 /* per-cpu buffer info */
 static struct buf_status
 {
-	struct buf_info info;
+	struct _stp_buf_info info;
 	unsigned max_backlog; /* max # sub-buffers ready at one time */
 } status[NR_CPUS];
 
+
+
 /**
  *	streaming - is the current transport mode streaming or not?
  *
@@ -140,7 +133,6 @@ int send_request(int type, void *data, int len)
 }
 
 
-
 /**
  *	summarize - print a summary if applicable
  */
@@ -294,7 +286,7 @@ static void wait_for_percpu_threads(int n)
 /**
  *	process_subbufs - write ready subbufs to disk
  */
-static int process_subbufs(struct buf_info *info)
+static int process_subbufs(struct _stp_buf_info *info)
 {
 	unsigned subbufs_ready, start_subbuf, end_subbuf, subbuf_idx, i;
 	int len, cpu = info->cpu;
@@ -332,7 +324,7 @@ static void *reader_thread(void *data)
 	int rc;
 	int cpu = (long)data;
 	struct pollfd pollfd;
-	struct consumed_info consumed_info;
+	struct _stp_consumed_info consumed_info;
 	unsigned subbufs_consumed;
 
 	pollfd.fd = relay_file[cpu];
@@ -352,7 +344,7 @@ static void *reader_thread(void *data)
 		}
 
 		rc = read(proc_file[cpu], &status[cpu].info,
-			  sizeof(struct buf_info));
+			  sizeof(struct _stp_buf_info));
 		subbufs_consumed = process_subbufs(&status[cpu].info);
 		if (subbufs_consumed) {
 			if (subbufs_consumed > status[cpu].max_backlog)
@@ -360,7 +352,7 @@ static void *reader_thread(void *data)
 			status[cpu].info.consumed += subbufs_consumed;
 			consumed_info.cpu = cpu;
 			consumed_info.consumed = subbufs_consumed;
-			if (write (proc_file[cpu], &consumed_info, sizeof(struct consumed_info)) < 0)
+			if (write (proc_file[cpu], &consumed_info, sizeof(struct _stp_consumed_info)) < 0)
 				fprintf(stderr,"WARNING: writing consumed info failed.\n");
 		}
 		if (status[cpu].info.flushing)
@@ -486,9 +478,6 @@ void system_cmd(char *cmd)
 	}
 }
 
-#include <sys/wait.h>
-static void cleanup_and_exit (int);
-
 /**
  *	init_stp - initialize the app
  *	@print_summary: boolean, print summary or not at end of run
@@ -498,7 +487,7 @@ static void cleanup_and_exit (int);
 int init_stp(int print_summary)
 {
 	char buf[1024];
-	struct transport_info ti;
+	struct _stp_transport_info ti;
 	pid_t pid;
 	int rstatus;
 
@@ -559,7 +548,7 @@ int init_stp(int print_summary)
 	}
 	return 0;
 
-do_rmmod:
+ do_rmmod:
 	snprintf(buf, sizeof(buf), "/sbin/rmmod -w %s", modname);
 	if (system(buf))
 		fprintf(stderr, "ERROR: couldn't rmmod probe module %s.\n", modname);
@@ -654,7 +643,7 @@ static int merge_output(void)
 	return 0;
 }
 
-static void cleanup_and_exit (int closed)
+void cleanup_and_exit (int closed)
 {
 	char tmpbuf[128];
 	pid_t err;
@@ -733,7 +722,6 @@ static char recvbuf[8192];
 int stp_main_loop(void)
 {
 	int nb, rc;
-	struct transport_start ts;
 	void *data;
 	int type;
 	FILE *ofp = stdout;
@@ -767,9 +755,41 @@ int stp_main_loop(void)
 		}
 
 		switch (type) {
+		case STP_REALTIME_DATA:
+			fwrite_unlocked(data, nb - sizeof(int), 1, ofp);
+			break;
+		case STP_OOB_DATA:
+			fputs ((char *)data, stderr);
+			break;
+		case STP_EXIT: 
+		{
+			/* module asks us to unload it and exit */
+			int *closed = (int *)data;
+			cleanup_and_exit(*closed);
+			break;
+		}
+		case STP_START: 
+		{
+			struct _stp_transport_start *t = (struct _stp_transport_start *)data;
+			dbug("probe_start() returned %d\n", t->pid);
+			if (t->pid < 0) {
+				if (target_cmd)
+					kill (target_pid, SIGKILL);
+				cleanup_and_exit(0);
+			} else if (target_cmd)
+				kill (target_pid, SIGUSR1);
+			break;
+		}
+		case STP_SYSTEM:
+		{
+			struct _stp_cmd_info *c = (struct _stp_cmd_info *)data;
+			system_cmd(c->cmd);
+			break;
+		}
 		case STP_TRANSPORT_INFO:
 		{
-			struct transport_info *info = (struct transport_info *)data;			
+			struct _stp_transport_info *info = (struct _stp_transport_info *)data;
+			struct _stp_transport_start ts;
 			transport_mode = info->transport_mode;
 			params.subbuf_size = info->subbuf_size;
 			params.n_subbufs = info->n_subbufs;
@@ -796,43 +816,39 @@ int stp_main_loop(void)
 				if (!ofp) {
 					fprintf (stderr, "ERROR: couldn't open output file %s: errcode = %s\n",
 						 outfile_name, strerror(errno));
-					/* FIXME. Need to cleanup properly */
-					exit(1);
+					cleanup_and_exit(0);
 				}
 			}
 			ts.pid = getpid();
 			send_request(STP_START, &ts, sizeof(ts));
 			break;
 		}
-		case STP_REALTIME_DATA:
-			fwrite_unlocked(data, nb - sizeof(int), 1, ofp);
-			break;
-		case STP_OOB_DATA:
-			fputs ((char *)data, stderr);
-			break;
-		case STP_EXIT: 
+		case STP_MODULE:
 		{
-			/* module asks us to unload it and exit */
-			int *closed = (int *)data;
-			cleanup_and_exit(*closed);
+			struct _stp_transport_start ts;
+			if (do_module(data)) {
+			  ts.pid = getpid();
+			  send_request(STP_START, &ts, sizeof(ts));
+			}
 			break;
-		}
-		case STP_START: 
+		}		
+		case STP_SYMBOLS:
 		{
-			struct transport_start *t = (struct transport_start *)data;
-			dbug("probe_start() returned %d\n", t->pid);
-			if (t->pid < 0) {
-				if (target_cmd)
-					kill (target_pid, SIGKILL);
-				cleanup_and_exit(0);
-			} else if (target_cmd)
-				kill (target_pid, SIGUSR1);
-			break;
-		}
-		case STP_SYSTEM:
-		{
-			struct cmd_info *c = (struct cmd_info *)data;
-			system_cmd(c->cmd);
+			struct _stp_symbol_req *req = (struct _stp_symbol_req *)data;
+			struct _stp_transport_start ts;
+			dbug("STP_SYMBOLS request received\n");
+			if (req->endian != 0x1234) {
+			  fprintf(stderr,"ERROR: staprun is compiled with different endianess than the kernel!\n");
+			  cleanup_and_exit(0);
+			}
+			if (req->ptr_size != sizeof(char *)) {
+			  fprintf(stderr,"ERROR: staprun is compiled with %d-bit pointers and the kernel uses %d-bit.\n",
+				  8*sizeof(char *), 8*req->ptr_size);
+			  cleanup_and_exit(0);
+			}
+			do_kernel_symbols();
+			ts.pid = getpid();
+			send_request(STP_START, &ts, sizeof(ts));
 			break;
 		}
 		default:
