@@ -20,14 +20,8 @@
  *
  */
 
-#include <stdio.h>
-#include <stdlib.h>
-#include <unistd.h>
-#include <string.h>
-#include <sys/wait.h>
-#include <sys/statfs.h>
-#include <pwd.h>
 #include "librelay.h"
+#include <pwd.h>
 
 extern char *optarg;
 extern int optopt;
@@ -35,9 +29,7 @@ extern int optind;
 
 int print_only = 0;
 int quiet = 0;
-int merge = 1;
 int verbose = 0;
-int enable_relayfs = 1;
 int target_pid = 0;
 int driver_pid = 0;
 unsigned int buffer_size = 0;
@@ -51,29 +43,13 @@ char *username = NULL;
 uid_t cmd_uid;
 gid_t cmd_gid;
 
- /* relayfs base file name */
-static char stpd_filebase[1024];
-#define RELAYFS_MAGIC			0xF0B4A981
-
-/* if no output file name is specified, use this */
-#define DEFAULT_OUTFILE_NAME "probe.out"
-
-/* stp_check script */
-#ifdef PKGLIBDIR
-char *stp_check=PKGLIBDIR "/stp_check";
-#else
-char *stp_check="stp_check";
-#endif
-
 static void usage(char *prog)
 {
 	fprintf(stderr, "\n%s [-m] [-p] [-q] [-r] [-c cmd ] [-t pid]\n"
                 "\t[-b bufsize] [-o FILE] kmod-name [kmod-options]\n", prog);
-	fprintf(stderr, "-m  Don't merge per-cpu files.\n");
 	fprintf(stderr, "-p  Print only.  Don't log to files.\n");
 	fprintf(stderr, "-q  Quiet. Don't display trace to stdout.\n");
-	fprintf(stderr, "-r  Disable running stp_check and loading relayfs module.\n");
-	fprintf(stderr, "-c cmd.  Command \'cmd\' will be run and stpd will exit when it does.\n");
+	fprintf(stderr, "-c cmd.  Command \'cmd\' will be run and staprun will exit when it does.\n");
 	fprintf(stderr, "   _stp_target will contain the pid for the command.\n");
 	fprintf(stderr, "-t pid.  Sets _stp_target to pid.\n");
 	fprintf(stderr, "-d pid.  Pass the systemtap driver's pid.\n");
@@ -88,15 +64,13 @@ static void usage(char *prog)
 
 int main(int argc, char **argv)
 {
-	int c, status;
-	pid_t pid;
-	struct statfs st;
-
+	int c;
+	
 	while ((c = getopt(argc, argv, "mpqrb:n:t:d:c:vo:u:")) != EOF)
 	{
 		switch (c) {
 		case 'm':
-			merge = 0;
+			fprintf(stderr, "Warning: -m option deprecated. Ignoring...\n");
 			break;
 		case 'p':
 			print_only = 1;
@@ -108,7 +82,7 @@ int main(int argc, char **argv)
 			verbose = 1;
 			break;
 		case 'r':
-			enable_relayfs = 0;
+			fprintf(stderr, "Warning: -r option deprecated. Ignoring...\n");
 			break;
 		case 'b':
 		{
@@ -191,36 +165,8 @@ int main(int argc, char **argv)
 		cmd_gid = getgid();
 	}
 
-	if (enable_relayfs) {
-		/* now run the _stp_check script */
-		if ((pid = fork()) < 0) {
-			perror ("fork of stp_check failed.");
-			exit(-1);
-		} else if (pid == 0) {
-			if (execlp(stp_check, stp_check, NULL) < 0)
-				_exit (-1);
-		}
-		if (waitpid(pid, &status, 0) < 0) {
-			perror("waitpid");
-			exit(-1);
-		}
-		if (WIFEXITED(status) && WEXITSTATUS(status)) {
-			perror (stp_check);
-			fprintf(stderr, "Could not execute %s\n", stp_check);
-			exit(1);
-		}
-		if (!outfile_name)
-			outfile_name = DEFAULT_OUTFILE_NAME;
-	}
-
-	if (statfs("/mnt/relay", &st) == 0
-	    && (int) st.f_type == (int) RELAYFS_MAGIC)
-		sprintf(stpd_filebase, "/mnt/relay/%d/cpu", getpid());
-	else
-		sprintf(stpd_filebase, "/proc/systemtap/stap_%d/cpu", driver_pid);
-
-	if (init_stp(stpd_filebase, !quiet)) {
-		//fprintf(stderr, "Couldn't initialize stpd. Exiting.\n");
+	if (init_stp(!quiet)) {
+		//fprintf(stderr, "Couldn't initialize staprun. Exiting.\n");
 		exit(1);
 	}
 
