@@ -18,12 +18,13 @@ spinlock_t _stp_pool_lock = SPIN_LOCK_UNLOCKED;
 spinlock_t _stp_ready_lock = SPIN_LOCK_UNLOCKED;
 
 #ifdef STP_RELAYFS
+extern int _stp_relay_flushing;
 /* handle the per-cpu subbuf info read for relayfs */
 static ssize_t
 _stp_proc_read (struct file *file, char __user *buf, size_t count, loff_t *ppos)
 {
 	int num;
-	struct buf_info out;
+	struct _stp_buf_info out;
 
 	int cpu = *(int *)(PDE(file->f_dentry->d_inode)->data);
 
@@ -38,6 +39,7 @@ _stp_proc_read (struct file *file, char __user *buf, size_t count, loff_t *ppos)
 	out.produced = atomic_read(&_stp_chan->buf[cpu]->subbufs_produced);
 	out.consumed = atomic_read(&_stp_chan->buf[cpu]->subbufs_consumed);
 #endif  /* RELAYFS_CHANNEL_VERSION >= 4 || CONFIG_RELAY */
+	out.flushing = _stp_relay_flushing;
 
 	num = sizeof(out);
 	if (copy_to_user(buf, &out, num))
@@ -50,7 +52,7 @@ _stp_proc_read (struct file *file, char __user *buf, size_t count, loff_t *ppos)
 static ssize_t _stp_proc_write (struct file *file, const char __user *buf,
 				size_t count, loff_t *ppos)
 {
-	struct consumed_info info;
+	struct _stp_consumed_info info;
 	int cpu = *(int *)(PDE(file->f_dentry->d_inode)->data);
 	if (copy_from_user(&info, buf, count))
 		return -EFAULT;
@@ -79,29 +81,42 @@ static ssize_t _stp_proc_write_cmd (struct file *file, const char __user *buf,
 
 	//printk ("_stp_proc_write_cmd. count:%d type:%d\n", count, type);
 
-	count -= sizeof(int);
+	if (type == STP_SYMBOLS) {
+		count -= sizeof(long);
+		buf += sizeof(long);
+	} else {
+		count -= sizeof(int);
+		buf += sizeof(int);
+	}
 
 	switch (type) {
 	case STP_START:
 	{
-		struct transport_start st;
-		if (count < sizeof(struct transport_start))
+		struct _stp_transport_start st;
+		if (count < sizeof(struct _stp_transport_start))
 			return 0;
-		if (copy_from_user (&st, buf, sizeof(struct transport_start)))
+		if (copy_from_user (&st, buf, sizeof(struct _stp_transport_start)))
 			return -EFAULT;
 		_stp_handle_start (&st);
 		break;
 	}
+
+	case STP_SYMBOLS:
+		count = _stp_do_symbols(buf, count);
+		break;
+	case STP_MODULE:
+		count = _stp_do_module(buf, count);
+		break;
 	case STP_EXIT:
 		_stp_exit_flag = 1;
 		break;
 	case STP_TRANSPORT_INFO:
 	{
-		struct transport_info ti;
-		//printk("STP_TRANSPORT_INFO %d %d\n", count, sizeof(struct transport_info));
-		if (count < sizeof(struct transport_info))
+		struct _stp_transport_info ti;
+		kbug("STP_TRANSPORT_INFO %d %d\n", (int)count, (int)sizeof(struct _stp_transport_info));
+		if (count < sizeof(struct _stp_transport_info))
 			return 0;
-		if (copy_from_user (&ti, &buf[4], sizeof(struct transport_info)))
+		if (copy_from_user (&ti, buf, sizeof(struct _stp_transport_info)))
 			return -EFAULT;
 		if (_stp_transport_open (&ti) < 0)
 			return -1;
@@ -128,10 +143,17 @@ static int _stp_write (int type, void *data, int len)
 {
 	struct _stp_buffer *bptr;
 	unsigned long flags;
+	unsigned numtrylock;
 
 #define WRITE_AGG
 #ifdef WRITE_AGG
-	spin_lock_irqsave(&_stp_ready_lock, flags);
+
+	numtrylock = 0;
+	while (!spin_trylock_irqsave (&_stp_ready_lock, flags) && (++numtrylock < MAXTRYLOCK)) 
+		ndelay (TRYLOCKDELAY);
+	if (unlikely (numtrylock >= MAXTRYLOCK))
+		return 0;
+
 	if (!list_empty(&_stp_ready_q)) {
 		bptr = (struct _stp_buffer *)_stp_ready_q.prev;
 		if (bptr->len + len <= STP_BUFFER_SIZE 
@@ -146,7 +168,12 @@ static int _stp_write (int type, void *data, int len)
 	spin_unlock_irqrestore(&_stp_ready_lock, flags);
 #endif
 
-	spin_lock_irqsave(&_stp_pool_lock, flags);
+	numtrylock = 0;
+	while (!spin_trylock_irqsave (&_stp_pool_lock, flags) && (++numtrylock < MAXTRYLOCK)) 
+		ndelay (TRYLOCKDELAY);
+	if (unlikely (numtrylock >= MAXTRYLOCK))
+		return 0;
+
 	if (list_empty(&_stp_pool_q)) {
 		spin_unlock_irqrestore(&_stp_pool_lock, flags);
 		return -1;
@@ -162,7 +189,11 @@ static int _stp_write (int type, void *data, int len)
 	bptr->len = len;
 	
 	/* put it on the pool of ready buffers */
-	spin_lock_irqsave(&_stp_ready_lock, flags);
+	numtrylock = 0;
+	while (!spin_trylock_irqsave (&_stp_ready_lock, flags) && (++numtrylock < MAXTRYLOCK)) 
+		ndelay (TRYLOCKDELAY);
+	if (unlikely (numtrylock >= MAXTRYLOCK))
+		return 0;
 	list_add_tail(&bptr->list, &_stp_ready_q);
 	spin_unlock_irqrestore(&_stp_ready_lock, flags);
 
@@ -244,7 +275,7 @@ static int _stp_set_buffers(int num)
 	
 	if (num > _stp_current_buffers) {
 		for (i = 0; i < num - _stp_current_buffers; i++) {
-			p = (struct list_head *)kmalloc(sizeof(struct _stp_buffer),GFP_KERNEL);
+			p = (struct list_head *)kmalloc(sizeof(struct _stp_buffer),STP_ALLOC_FLAGS);
 			if (!p)	{
 				_stp_current_buffers += i;
 				goto err;
@@ -267,38 +298,6 @@ err:
 	return _stp_current_buffers;
 }
 
-#if defined (STP_RELAYFS) && defined(CONFIG_RELAY)
-struct dentry *module_dir_dentry;
-
-static inline struct dentry *_stp_get_proc_root(void)
-{
-	struct file_system_type *procfs_type;
-	struct super_block *procfs_sb;
-
-	procfs_type = get_fs_type("proc");
-	if (!procfs_type || list_empty(&procfs_type->fs_supers))
-		return NULL;
-	procfs_sb = list_entry(procfs_type->fs_supers.next,
-			       struct super_block, s_instances);
-	return procfs_sb->s_root;
-}
-
-static inline struct dentry *_stp_force_dir_creation(const char *dirname, struct dentry *parent)
-{
-	struct dentry *dir_dentry;
-
-	mutex_lock(&parent->d_inode->i_mutex);
-	dir_dentry = lookup_one_len(dirname, parent, strlen(dirname));
-	mutex_unlock(&parent->d_inode->i_mutex);
-	if (IS_ERR(dir_dentry)) {
-		dir_dentry = NULL;
-		remove_proc_entry(dirname, NULL);
-	}
-	
-	return dir_dentry;
-}
-#endif  /* STP_RELAYFS && CONFIG_RELAY */
-
 static int _stp_register_procfs (void)
 {
 	int i;
@@ -307,10 +306,6 @@ static int _stp_register_procfs (void)
 	int j;
 	char buf[8];
 #endif
-#if defined (CONFIG_RELAY)
-	struct dentry *proc_root_dentry;
-	struct dentry *systemtap_dir_dentry;
-#endif /* CONFIG_RELAY */
 	struct proc_dir_entry *de;
 	struct list_head *p, *tmp;
 
@@ -319,7 +314,7 @@ static int _stp_register_procfs (void)
 
 	/* allocate buffers */
 	for (i = 0; i < STP_DEFAULT_BUFFERS; i++) {
-		p = (struct list_head *)kmalloc(sizeof(struct _stp_buffer),GFP_KERNEL);
+		p = (struct list_head *)kmalloc(sizeof(struct _stp_buffer),STP_ALLOC_FLAGS);
 		// printk("allocated buffer at %lx\n", (long)p);
 		if (!p)
 			goto err2;
@@ -342,23 +337,11 @@ static int _stp_register_procfs (void)
 			goto err0;
 	}
 
-#if defined (STP_RELAYFS) && defined (CONFIG_RELAY)
-	proc_root_dentry = _stp_get_proc_root();
-	systemtap_dir_dentry = _stp_force_dir_creation(dirname, proc_root_dentry);
-	if (!systemtap_dir_dentry)
-		goto err0;
-#endif /* STP_RELAYFS && CONFIG_RELAY */
 	/* now create /proc/systemtap/module_name */
 	_stp_proc_mod = proc_mkdir (THIS_MODULE->name, _stp_proc_root);
 	if (_stp_proc_mod == NULL)
 		goto err0;
 
-#if defined (STP_RELAYFS) && defined (CONFIG_RELAY)
-	module_dir_dentry = _stp_force_dir_creation(THIS_MODULE->name, systemtap_dir_dentry);
-	if (!module_dir_dentry)
-		goto err0;
-#endif /* STP_RELAYFS && CONFIG_RELAY */
-
 #ifdef STP_RELAYFS	
 	/* now for each cpu "n", create /proc/systemtap/module_name/n  */
 	for_each_cpu(i) {
@@ -367,7 +350,7 @@ static int _stp_register_procfs (void)
 		if (de == NULL) 
 			goto err1;
 		de->proc_fops = &_stp_proc_fops;
-		de->data = kmalloc(sizeof(int), GFP_KERNEL);
+		de->data = kmalloc(sizeof(int), STP_ALLOC_FLAGS);
 		if (de->data == NULL) {
 			remove_proc_entry (buf, _stp_proc_mod);
 			goto err1;
