@@ -1,5 +1,5 @@
 // -*- C++ -*-
-// Copyright (C) 2005 Red Hat Inc.
+// Copyright (C) 2005, 2006 Red Hat Inc.
 //
 // This file is part of systemtap, and is free software.  You can
 // redistribute it and/or modify it under the terms of the GNU General
@@ -9,12 +9,12 @@
 #ifndef SESSION_H
 #define SESSION_H
 
-#include "elaborate.h"
 #include <string>
 #include <vector>
 #include <iostream>
 #include <sstream>
 #include <map>
+#include <set>
 
 extern "C" {
 #include <elfutils/libdw.h>
@@ -26,7 +26,14 @@ struct match_node;
 struct stapfile;
 struct vardecl;
 struct functiondecl;
-struct derived_probe_group_container;
+struct derived_probe;
+struct be_derived_probe_group;
+struct dwarf_derived_probe_group;
+struct timer_derived_probe_group;
+struct profile_derived_probe_group;
+struct mark_derived_probe_group;
+struct hrtimer_derived_probe_group;
+struct perfmon_derived_probe_group;
 struct embeddedcode;
 struct translator_output;
 struct unparser;
@@ -61,6 +68,7 @@ struct statistic_decl
 struct systemtap_session
 {
   systemtap_session ();
+  // NB: new POD members likely need to be explicitly cleared in the ctor.
 
   // command line args
   std::vector<std::string> include_path;
@@ -70,6 +78,7 @@ struct systemtap_session
   std::string kernel_base_release;
   std::string architecture;
   std::string runtime_path;
+  std::string data_path;
   std::string module_name;
   std::string output_file;
   std::string cmd;
@@ -84,6 +93,12 @@ struct systemtap_session
   bool merge;
   int buffer_size;
   unsigned perfmon;
+  bool symtab;
+
+  // Cache data
+  bool use_cache;
+  std::string cache_path;
+  std::string hash_path;
 
   // temporary directory for module builds etc.
   // hazardous - it is "rm -rf"'d at exit
@@ -101,13 +116,20 @@ struct systemtap_session
   std::vector<stapfile*> files;
   std::vector<vardecl*> globals;
   std::vector<functiondecl*> functions;
-  derived_probe_group_container probes;
+  std::vector<derived_probe*> probes; // see also *_probes groups below
   std::vector<embeddedcode*> embeds;
   std::map<std::string, statistic_decl> stat_decls;
   // XXX: vector<*> instead please?
 
-  // module-referencing file handles
-  std::map<std::string,int> module_fds;
+  // Every probe in these groups must also appear in the
+  // session.probes vector.
+  be_derived_probe_group* be_derived_probes;
+  dwarf_derived_probe_group* dwarf_derived_probes;
+  timer_derived_probe_group* timer_derived_probes;
+  profile_derived_probe_group* profile_derived_probes;
+  mark_derived_probe_group* mark_derived_probes;
+  hrtimer_derived_probe_group* hrtimer_derived_probes;
+  perfmon_derived_probe_group* perfmon_derived_probes;
 
   // unparser data
   translator_output* op;
@@ -118,9 +140,12 @@ struct systemtap_session
   Dwarf_Addr kprobes_text_start;
   Dwarf_Addr kprobes_text_end;
 
-  unsigned num_errors;
+  std::set<std::string> seen_errors;
+  unsigned num_errors () { return seen_errors.size(); }
   // void print_error (const parse_error& e);
   void print_error (const semantic_error& e);
+
+  // reNB: new POD members likely need to be explicitly cleared in the ctor.
 };
 
 
