@@ -9,6 +9,8 @@
 #include "config.h"
 #include "staptree.h"
 #include "parse.h"
+#include "util.h"
+
 #include <iostream>
 #include <typeinfo>
 #include <sstream>
@@ -17,37 +19,6 @@
 using namespace std;
 
 
-// return as quoted string, with at least '"' backslash-escaped
-template <typename IN> inline string
-lex_cast_qstring(IN const & in)
-{
-  stringstream ss;
-  string out, out2;
-  if (!(ss << in))
-    throw runtime_error("bad lexical cast");
-  out = ss.str();
-  out2 += '"';
-  for (unsigned i=0; i<out.length(); i++)
-    {
-      if (out[i] == '"') // XXX others?
-	out2 += '\\';
-      out2 += out[i];
-    }
-  out2 += '"';
-  return out2;
-}
-
-
-template <typename OUT, typename IN> inline OUT
-lex_cast(IN const & in)
-{
-  stringstream ss;
-  OUT out;
-  if (!(ss << in && ss >> out))
-    throw runtime_error("bad lexical cast");
-  return out;
-}
-
 
 expression::expression ():
   type (pe_unknown), tok (0)
@@ -133,7 +104,7 @@ probe_point::component::component (std::string const & f, literal * a):
 
 
 vardecl::vardecl ():
-  arity (-1)
+  arity (-1), maxsize(0), init(NULL)
 {
 }
 
@@ -144,7 +115,7 @@ vardecl::set_arity (int a)
   if (a < 0)
     return;
 
-  if (arity != a && arity >= 0)
+  if ((arity != a && arity >= 0) || (a == 0 && maxsize > 0))
     throw semantic_error ("inconsistent arity", tok);
 
   if (arity != a)
@@ -159,6 +130,8 @@ vardecl::set_arity (int a)
 bool 
 vardecl::compatible_arity (int a)
 {
+  if (a == 0 && maxsize > 0)
+    return false;
   if (arity == -1 || a == -1)
     return true;
   return arity == a;
@@ -294,14 +267,24 @@ void target_symbol::print (std::ostream& o) const
 void vardecl::print (ostream& o) const
 {
   o << name;
+  if (maxsize > 0)
+    o << "[" << maxsize << "]";
   if (arity > 0 || index_types.size() > 0)
     o << "[...]";
+  if (init)
+    {
+      o << " = ";
+      init->print(o);
+    }
 }
 
 
 void vardecl::printsig (ostream& o) const
 {
-  o << name << ":" << type;
+  o << name;
+  if (maxsize > 0)
+    o << "[" << maxsize << "]";
+  o << ":" << type;
   if (index_types.size() > 0)
     {
       o << " [";
@@ -370,8 +353,11 @@ print_format::components_to_string(vector<format_component> const & components)
 	  for (string::const_iterator j = i->literal_string.begin();
 	       j != i->literal_string.end(); ++j)
 	    {
+              // See also: c_unparser::visit_literal_string and lex_cast_qstring
 	      if (*j == '%')
 		oss << '%';
+              else if(*j == '"')
+                oss << '\\';
 	      oss << *j;
 	    }
 	}
@@ -752,6 +738,11 @@ void foreach_loop::print (ostream& o) const
   base->print_indexable (o);
   if (sort_direction != 0 && sort_column == 0)
     o << (sort_direction > 0 ? "+" : "-");
+  if (limit)
+    {
+      o << " limit ";
+      limit->print (o);
+    }
   o << ") ";
   block->print (o);
 }
@@ -1320,6 +1311,10 @@ traversing_visitor::visit_foreach_loop (foreach_loop* s)
 
   for (unsigned i=0; i<s->indexes.size(); i++)
     s->indexes[i]->visit (this);
+
+  if (s->limit)
+    s->limit->visit (this);
+
   s->block->visit (this);
 }
 
@@ -1965,6 +1960,7 @@ deep_copy_visitor::visit_foreach_loop (foreach_loop* s)
 
   n->sort_direction = s->sort_direction;
   n->sort_column = s->sort_column;
+  require <expression*> (this, &(n->limit), s->limit);
 
   require <statement*> (this, &(n->block), s->block);
   provide <foreach_loop*> (this, n);
