@@ -1,5 +1,6 @@
 // -*- C++ -*-
-// Copyright (C) 2005 Red Hat Inc.
+// Copyright (C) 2005, 2006 Red Hat Inc.
+// Copyright (C) 2006 Intel Corporation.
 //
 // This file is part of systemtap, and is free software.  You can
 // redistribute it and/or modify it under the terms of the GNU General
@@ -9,7 +10,6 @@
 #ifndef STAPTREE_H
 #define STAPTREE_H
 
-#include "session.h"
 #include <map>
 #include <stack>
 #include <set>
@@ -26,17 +26,18 @@ struct token; // parse.h
 struct semantic_error: public std::runtime_error
 {
   const token* tok1;
-  const std::string msg2;
+  std::string msg2;
   const token* tok2;
+  semantic_error *chain;
 
   ~semantic_error () throw () {}
   semantic_error (const std::string& msg):
-    runtime_error (msg), tok1 (0), tok2 (0) {}
+    runtime_error (msg), tok1 (0), tok2 (0), chain (0) {}
   semantic_error (const std::string& msg, const token* t1):
-    runtime_error (msg), tok1 (t1), tok2 (0) {}
+    runtime_error (msg), tok1 (t1), tok2 (0), chain (0) {}
   semantic_error (const std::string& msg, const token* t1,
                   const std::string& m2, const token* t2):
-    runtime_error (msg), tok1 (t1), msg2 (m2), tok2 (t2) {}
+    runtime_error (msg), tok1 (t1), msg2 (m2), tok2 (t2), chain (0) {}
 };
 
 // ------------------------------------------------------------------------
@@ -230,6 +231,8 @@ struct target_symbol : public expression
     };
   std::string base_name;
   std::vector<std::pair<component_type, std::string> > components;
+  semantic_error* saved_conversion_error;
+  target_symbol(): saved_conversion_error (0) {}
   void print (std::ostream& o) const;
   void visit (visitor* u);
 };
@@ -392,7 +395,9 @@ struct vardecl: public symboldecl
   void set_arity (int arity);
   bool compatible_arity (int a);
   int arity; // -1: unknown; 0: scalar; >0: array
+  int maxsize; // upperbound on size for arrays
   std::vector<exp_type> index_types; // for arrays only
+  literal *init; // for global scalars only
 };
 
 
@@ -463,6 +468,7 @@ struct foreach_loop: public statement
   indexable *base;
   int sort_direction; // -1: decreasing, 0: none, 1: increasing
   unsigned sort_column; // 0: value, 1..N: index
+  expression* limit; // optional iteration limit
 
   statement* block;
   void print (std::ostream& o) const;
